# Snake in napari
# ---------------
#
# Player 1 keys:
player1_up_key = 'w'
player1_down_key = 's'
player1_left_key = 'a'
player1_right_key = 'd'

# Player 1 keys:
player2_up_key = 'i'
player2_down_key = 'k'
player2_left_key = 'j'
player2_right_key = 'l'

#
# Have fun!
#   @haesleinhuepf

import time
import napari
import numpy as np
from napari.qt.threading import thread_worker
from qtpy.QtWidgets import QLineEdit, QLabel, QWidget, QVBoxLayout
from ._utils import draw_box
from scipy.ndimage import maximum_filter
from napari_tools_menu import register_action

class Game:

    def __init__(self):
        """ Setup the game
        """

        # playground config
        self.width = 640
        self.height = 480

        self.pixel_size = 10
        self.food_calories = 5
        self.maximum_food_available = 10

        self.frame_delay = 0.2 # seconds

        # player 1
        self.player1_delta_x = 0
        self.player1_delta_y = self.pixel_size
        self.player1_score = 0
        self.player1_positions = [
            [240, 240]
        ]

        # player 2
        self.player2_delta_x = 0
        self.player2_delta_y = -self.pixel_size
        self.player2_score = 0
        self.player2_positions = [
            [480, 240]
        ]

        self.food_positions = []

        # others
        self.iteration = 0

        self.playground = np.zeros([self.height, self.width])
        self.temp = np.zeros([self.height, self.width])

        self.result_label = QLabel()

    def set_player1_direction(self, delta_x, delta_y):
        self.player1_delta_x = delta_x * self.pixel_size
        self.player1_delta_y = delta_y * self.pixel_size

    def set_player2_direction(self, delta_x, delta_y):
        self.player2_delta_x = delta_x * self.pixel_size
        self.player2_delta_y = delta_y * self.pixel_size

    def game_step(self):
        """Forwards the game by one step and computes the new playground

        Returns
        -------
            an image with the current state of the game
        """

        self.result_label.setText(str(self.player1_score) + " : " + str(self.player2_score))

        # move player 1
        result = self.move_player(self.player1_positions, self.player1_delta_x, self.player1_delta_y, self.player1_score)
        if result is None:
            return self.playground
        self.player1_positions = result

        # move player 2
        result = self.move_player(self.player2_positions, self.player2_delta_x, self.player2_delta_y, self.player2_score)
        if result is None:
            return self.playground
        self.player2_positions = result

        # go through food positions and check if a player ate it
        new_food_positions = []
        for pos in self.food_positions:
            if pos == self.player1_positions[0]:
                self.player1_score = self.player1_score + self.food_calories
            elif pos == self.player2_positions[0]:
                self.player2_score = self.player2_score + self.food_calories
            else:
                new_food_positions.append(pos)
        self.food_positions = new_food_positions

        # seed new food from time to time
        if len(self.food_positions) < self.maximum_food_available:
            random_x = (int(np.random.random_sample() * self.width / self.pixel_size - 2) + 1) * self.pixel_size
            random_y = (int(np.random.random_sample() * self.height / self.pixel_size - 2) + 1) * self.pixel_size
            if [random_x, random_y] not in self.player1_positions and \
                [random_x, random_y] not in self.player2_positions and \
                [random_x, random_y] not in self.food_positions:
                self.food_positions.append([random_x, random_y])

        # draw playground frame
        draw_box(self.temp, 0, 0, 0, self.width, self.height, 1, 4)
        draw_box(self.temp, 1, 1, 0, self.width - 3, self.height - 3, 1, 0)

        # draw players and food
        self.draw_positions(self.player1_positions, 2)
        self.draw_positions(self.player2_positions, 7)
        self.draw_positions(self.food_positions, 10)

        self.playground = maximum_filter(self.temp, size=self.pixel_size)

        # return playground
        image = self.playground

        return image

    def move_player(self, positions, player_delta_x, player_delta_y, player_score):
        """ Move a player by one step and check if it hit the wall, itself or the other player.
        The game is over then.
        """
        player_position_x = positions[0][0] + player_delta_y
        player_position_y = positions[0][1] + player_delta_x

        if self.is_game_over(player_position_x, player_position_y):
            print("Game over!")
            time.sleep(5)
            self.__init__()
            return None

        # move snake ahead
        new_positions = [
            [player_position_x, player_position_y]
        ]
        for i, pos in enumerate(positions):
            if i > player_score + 2:
                break
            new_positions.append(pos)


        return new_positions

    def is_game_over(self, player_position_x, player_position_y):
        return [player_position_x, player_position_y] in self.player1_positions or \
               [player_position_x, player_position_y] in self.player2_positions or \
               player_position_x <= 0 or \
               player_position_x >= self.width or \
               player_position_y <= 0 or \
               player_position_y >= self.height

    def draw_positions(self, new_positions, value):
        """bring position lists in the right format and draw the list of coordinates in a given color on the playground.
        """
        for p in new_positions:
            try:
                self.temp[(p[1], p[0])] = value
            except IndexError:
                pass

@register_action(menu="Games > Snake")
def snake(viewer : napari.Viewer):
    viewer.title = "natari"

    game = Game()

    # Key bindings for user control
    @viewer.bind_key(player1_up_key, overwrite=True)
    def player1_up_event(viewer):
        game.set_player1_direction(-1, 0)

    @viewer.bind_key(player1_down_key, overwrite=True)
    def player1_down_event(viewer):
        game.set_player1_direction(1, 0)

    @viewer.bind_key(player1_left_key, overwrite=True)
    def player1_left_event(viewer):
        game.set_player1_direction(0, -1)

    @viewer.bind_key(player1_right_key, overwrite=True)
    def player1_right_event(viewer):
        game.set_player1_direction(0, 1)

    # Key bindings for user control
    @viewer.bind_key(player2_up_key, overwrite=True)
    def player2_up_event(viewer):
        game.set_player2_direction(-1, 0)

    @viewer.bind_key(player2_down_key, overwrite=True)
    def player2_down_event(viewer):
        game.set_player2_direction(1, 0)

    @viewer.bind_key(player2_left_key, overwrite=True)
    def player2_left_event(viewer):
        game.set_player2_direction(0, -1)

    @viewer.bind_key(player2_right_key, overwrite=True)
    def player2_right_event(viewer):
        game.set_player2_direction(0, 1)

    # Graphical user interface
    widget = QWidget()
    layout = QVBoxLayout()
    widget.setLayout(layout)

    layout.addWidget(game.result_label)
    viewer.window.add_dock_widget(widget, area="bottom")
    viewer.window.add_dock_widget(widget, area="bottom")
    game.result_label.setText(str("0"))

    # Multi-threaded interaction
    # inspired by https://napari.org/docs/dev/events/threading.html
    def update_layer(new_image):
        try:
            viewer.layers['result'].data = new_image
        except KeyError:
            viewer.add_image(
                new_image, name='result', contrast_limits=(0, 10), colormap='turbo'
            )

    @thread_worker
    def yield_random_images_forever():
        while True:  # infinite loop!
            yield game.game_step()
            time.sleep(game.frame_delay)

    # Start the game loop
    worker = yield_random_images_forever()
    worker.yielded.connect(update_layer)
    worker.start()



"""
This module is an example of a barebones function plugin for napari

It implements the ``napari_experimental_provide_function`` hook specification.
see: https://napari.org/docs/dev/plugins/hook_specifications.html

Replace code below according to your needs.
"""
from typing import TYPE_CHECKING

from enum import Enum
import numpy as np
from napari_plugin_engine import napari_hook_implementation
from .cell_counting_arcade import cell_counting_arcade_with_default_image
from .ping_pong import ping_pong
from .snake import snake

if TYPE_CHECKING:
    import napari


# This is the actual plugin function, where we export our function
# (The functions themselves are defined below)
@napari_hook_implementation
def napari_experimental_provide_function():
    # we can return a single function
    # or a tuple of (function, magicgui_options)
    # or a list of multiple functions with or without options, as shown here:
    return [cell_counting_arcade_with_default_image, ping_pong, snake]


# 1.  First example, a simple function that thresholds an image and creates a labels layer
def threshold(data: "napari.types.ImageData", threshold: int) -> "napari.types.LabelsData":
    """Threshold an image and return a mask."""
    return (data > threshold).astype(int)


# 2. Second example, a function that adds, subtracts, multiplies, or divides two layers

# using Enums is a good way to get a dropdown menu.  Used here to select from np functions
class Operation(Enum):
    add = np.add
    subtract = np.subtract
    multiply = np.multiply
    divide = np.divide


def image_arithmetic(
    layerA: "napari.types.ImageData", operation: Operation, layerB: "napari.types.ImageData"
) -> "napari.types.LayerDataTuple":
    """Adds, subtracts, multiplies, or divides two same-shaped image layers."""
    return (operation.value(layerA, layerB), {"colormap": "turbo"})



# Ping-Pong in napari
# -------------------
#
# Player 1 keys:
player1_up_key = 'w'
player1_down_key = 's'
#
# Player 2 keys:
player2_up_key = 'i'
player2_down_key = 'k'
#
# Have fun!
#   @haesleinhuepf

import time
import napari
from napari.qt.threading import thread_worker
from qtpy.QtWidgets import QLineEdit, QLabel, QWidget, QVBoxLayout
import numpy as np
from ._utils import draw_box
from napari_tools_menu import register_action

class Game:

    def __init__(self):
        """ Setup the game
        """
        self.player1_position = 200
        self.player2_position = 280

        self.player1_x = 10
        self.player2_x = 630

        self.player1_score = 0
        self.player2_score = 0

        self.width = 640
        self.height = 480
        self.playground = np.zeros([self.height, self.width])

        self.bar_radius = 50

        self.puck_x = self.width / 2
        self.puck_y = self.height / 2
        self.puck_delta_x = 10
        self.puck_delta_y = 0

        self.result_label = QLabel()

    def game_step(self):
        """Forwards the game by one step and computes the new playground

        Returns
        -------
            an image with the current state of the game
        """

        # check player positions
        self.player1_position = self._check_player_position(self.player1_position)
        self.player2_position = self._check_player_position(self.player2_position)

        # move puck
        self.puck_x = self.puck_x + self.puck_delta_x
        self.puck_y = self.puck_y + self.puck_delta_y

        # check puck_position
        if self.puck_y < 0 or self.puck_y > self.height:
            self.puck_delta_y = -self.puck_delta_y
            self.puck_y = self.puck_y + self.puck_delta_y

        # puck at player 1
        if self.puck_x <= self.player1_x:
            self.puck_delta_x = -self.puck_delta_x
            self.puck_x = self.puck_x + self.puck_delta_x

            if abs(self.puck_y - self.player1_position) > self.bar_radius:
                # player 2 scores
                self.player2_score = self.player2_score + 1
                self._level_up()
            else:
                self.puck_delta_y = (self.puck_y - self.player1_position) / self.bar_radius * 5

        # puck at player 2
        if self.puck_x >= self.player2_x:
            self.puck_delta_x = -self.puck_delta_x
            self.puck_x = self.puck_x + self.puck_delta_x

            if abs(self.puck_y - self.player2_position) > self.bar_radius:
                # player 1 scores
                self.player1_score = self.player1_score + 1
                self._level_up()
            else:
                self.puck_delta_y = (self.puck_y - self.player2_position) / self.bar_radius * 5

        # show game status
        self.result_label.setText(str(self.player1_score) + ":" + str(self.player2_score))

        # draw playground
        self.playground.fill(0.1)

        # draw player 1
        draw_box(self.playground, self.player1_x, self.player1_position - self.bar_radius, 0, 10, self.bar_radius * 2, 0)

        # draw player 2
        draw_box(self.playground, self.player2_x, self.player2_position - self.bar_radius, 0, 10, self.bar_radius * 2, 0)

        # draw puck
        draw_box(self.playground, self.puck_x, self.puck_y, 0, 10, 5, 1)

        # return playground
        image = self.playground

        return image

    def _check_player_position(self, position):
        """Checks if a player went out of the playground

        Parameters
        ----------
        position: int
            current position of the player

        Returns
        -------
            new, potentially corrected position of the player
        """

        if position - self.bar_radius < 0:
            position = self.bar_radius
        if position + self.bar_radius > self.height:
            position = self.height - self.bar_radius
        return position

    def _level_up(self):
        """If a player scores, the game restarts with smaller bars or accelerated puck speed.

        Returns
        -------

        """
        if self.bar_radius > 10:
            self.bar_radius = self.bar_radius - 10
        else:
            self.puck_delta_x = self.puck_delta_x / abs(self.puck_delta_x) * (abs(self.puck_delta_x) + 10)

        self.puck_x = self.width / 2
        self.puck_y = self.height / 2

@register_action(menu="Games > Ping pong")
def ping_pong(viewer : napari.Viewer):
    viewer.title = "natari"

    game = Game()

    # Key bindings for user control
    @viewer.bind_key(player1_up_key)
    def player1_up_event(viewer):
        game.player1_position = game.player1_position - 10

    @viewer.bind_key(player1_down_key)
    def player1_down_event(viewer):
        game.player1_position = game.player1_position + 10

    @viewer.bind_key(player2_up_key)
    def player2_up_event(viewer):
        game.player2_position = game.player2_position - 10

    @viewer.bind_key(player2_down_key)
    def player2_down_event(viewer):
        game.player2_position = game.player2_position + 10

    # Graphical user interface
    widget = QWidget()
    layout = QVBoxLayout()
    widget.setLayout(layout)

    layout.addWidget(game.result_label)
    viewer.window.add_dock_widget(widget, area="bottom")
    game.result_label.setText(str("0:0"))

    # Multi-threaded interaction
    # inspired by https://napari.org/docs/dev/events/threading.html
    def update_layer(new_image):
        try:
            viewer.layers['result'].data = new_image
        except KeyError:
            viewer.add_image(
                new_image, name='result', contrast_limits=(0, 1)
            )

    @thread_worker
    def yield_random_images_forever():
        while True:  # infinite loop!
            yield game.game_step()
            time.sleep(0.05)

    # Start the game loop
    worker = yield_random_images_forever()
    worker.yielded.connect(update_layer)
    worker.start()



# Cell Counting Arcade
#
# This game runs in napari. It shows an image with nuclei and cells.
# It uses numpy to draw a playground where you can remove cells by
# shooting at them.
#
# Have fun!
#   @haesleinhuepf
#
# Installation: It's recommended to install these dependencies
# before running the game from the command line:
#
# pip install napari==0.4.11 natari
#
# Player keys:
player_left_key = "1"
player_right_key = "2"
player_fire_key = "9"
#
#
# We used image set BBBC022v1 [Gustafsdottir et al., PLOS ONE, 2013], available from the
# Broad Bioimage Benchmark Collection [Ljosa et al., Nature Methods, 2012].
from tifffile import imread

import time
import napari
from napari._qt.qthreading import thread_worker
from napari.types import LabelsData
import numpy as np
from pathlib import Path
from ._utils import draw_box
from napari_tools_menu import register_action

# The game
class CellCountingArcade():
    """
    The game allows the player to shoot bullets from the bottom of the screen which move up and if they hit a nucleus
    it is removed from the image data in the viewer with the surrounding cell.
    """
    def __init__(self, images, nuclei : LabelsData, cells : LabelsData, viewer : napari.Viewer):
        self.images = images
        self.nuclei = nuclei
        self.cells = cells
        self.viewer = viewer
        self.size = list(nuclei.shape)
        self.size[1] = int(0.9 * self.size[1])
        self.player_position = self.size[1] / 2
        self.bullets = []
        self.playground = np.zeros(nuclei.shape)
        self.fov_nuclei = np.zeros(nuclei.shape)
        self.fov_cells = np.zeros(nuclei.shape)

        self.fov_x = 0
        self.fov_delta_x = 1
        self.fov_max_x = nuclei.shape[1] - self.size[1]

    def move_player(self, delta):
        """
        Move the player left/right.
        """
        if self.player_position + delta > 0 and self.player_position + delta < self.size[1]:
            self.player_position += delta

    def fire(self):
        """
        Shoot a bullet
        """
        self.bullets.append([self.player_position, 0])

    def game_loop(self):
        """
        This function is called at every game iteration. It checks if bullets hit nuclei and redraws the playground
        """
        bullet_radius = 5

        # empty playground
        self.playground.fill(0)

        # pull nuclei from GPU so that we can access individual pixels
        nuclei = np.asarray(self.nuclei)
        # prepare all existing labels in a list. We set them to 0 in case a nucleus was hit
        labels_to_keep = list(range(0, int(np.max(self.nuclei)+1)))

        # future bullets to keep
        new_bullets = []

        for bullet in self.bullets:
            bullet[1] += 10

            # check if a bullet has hit a nucleus
            try:
                label = nuclei[int(self.playground.shape[0] - bullet[1]), int(bullet[0] + self.fov_x)]
            except IndexError:
                label = 0
            if label != 0: # bullet has hit a nucleus
                labels_to_keep[label] = 0
            elif bullet[1] > self.playground.shape[0]:
                pass # bullet has left the playground
            else:
                new_bullets.append(bullet)
                # draw bullets
                #self.playground[bullet[0]:bullet[0]+bullet_radius, self.playground.shape[0] - bullet[1] : self.playground.shape[0]].fill(1)
                draw_box(self.playground, bullet[0], self.playground.shape[0] - bullet[1], 0, bullet_radius, bullet_radius, 1, 1)
        self.bullets = new_bullets

        # only keep cells where the nuclei weren't hit
        new_label_ids = np.asarray(labels_to_keep, np.uint8)
        self.nuclei = np.take(new_label_ids, self.nuclei)
        self.cells = np.take(new_label_ids, self.cells)

        # make a binary image of areas to keep
        binary = self.cells > 0

        # draw player
        draw_box(self.playground, self.player_position - 5, self.playground.shape[0] - 20, 0, 10, 20, 1, 2)
        draw_box(self.playground, self.player_position - 15, self.playground.shape[0] - 10, 0, 30, 10, 1, 2)

        # collect all layers in a dictionary
        result = {}
        for i, image in enumerate(self.images):
            result["channel" + str(i)] = self.crop_fov(image * binary)

        # add segmentation (invisble) and playground
        result['nuclei'] = self.crop_fov(self.nuclei, self.fov_nuclei)
        result['cells'] = self.crop_fov(self.cells, self.fov_cells)
        result['playground'] = self.playground + 0

        self.fov_x += self.fov_delta_x
        if self.fov_x <= 0:
            self.fov_x = 0
            self.fov_delta_x = 1
        elif self.fov_x >= self.fov_max_x:
            self.fov_x = self.fov_max_x
            self.fov_delta_x = -1

        return result

    def crop_fov(self, image, output=None):
        return image[0:self.size[0], self.fov_x:self.fov_x+self.size[1]]



colours = ['magenta', 'green', 'cyan', 'gray']

@register_action(menu="Games > Cell counting arcade")
def cell_counting_arcade_with_default_image(viewer : napari.Viewer):
    viewer.title = "natari"

    images = []
    data_path = Path(__file__).parent / "data"
    dataset = imread(data_path / "IXMtest_A02_s9.tif")
    nuclei_channel = 0

    # add the original image channels as independent layers
    for i in range(0, dataset.shape[0]):
        viewer.add_image(dataset[i], blending='additive', name="channel" + str(i), colormap=colours[i])
        images.append(dataset[i])

    # image segmentation: nuclei and cells
    from skimage.filters import threshold_otsu
    from skimage.measure import label
    from skimage.segmentation import expand_labels

    binary_image = dataset[nuclei_channel] > threshold_otsu(dataset[nuclei_channel])
    labels_nuclei = label(binary_image)
    labels_cells = expand_labels(labels_nuclei, distance=50)

    cell_counting_arcade(viewer, np.asarray(labels_nuclei), np.asarray(labels_cells))

def cell_counting_arcade(viewer : napari.Viewer, labels_nuclei:LabelsData, labels_cells:LabelsData):

    images = []
    for l in viewer.layers:
        if isinstance(l, napari.layers.Image):
            images.append(l.data)

    game = CellCountingArcade(images, labels_nuclei, labels_cells, viewer)

    def update_layers(images_data: dict):
        """
        Add images to napari is layer or updates a pre-existing layer
        """
        for name in images_data.keys():
            image = images_data[name]
            for layer in viewer.layers:
                if layer.name == name:
                    layer.data = image
                    image = None
                    break

            if image is not None:
                if "nuclei" in name or "cells" in name:
                    viewer.add_labels(image, name=name, visible=False)
                else:
                    viewer.add_image(image, name=name, blending='additive')

    print("setting key bindings: ", player_left_key, player_right_key, player_fire_key)

    # Key bindings for the game
    @viewer.bind_key(player_left_key)
    def player_left_event(viewer):
        game.move_player(-10)

    @viewer.bind_key(player_right_key)
    def player_right_event(viewer):
        game.move_player(10)

    @viewer.bind_key(player_fire_key)
    def player_left_event(viewer):
        game.fire()

    print("Starting game loop")

    # Game loop, runs in the background
    # https://napari.org/guides/stable/threading.html
    @thread_worker
    def loop_run():
        while True:  # endless loop
            data = game.game_loop()
            yield data
            time.sleep(0.1)

    # Start the loop
    worker = loop_run()
    worker.yielded.connect(update_layers)
    worker.start()

    # napari.run()


